# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime
from random import choice, randint, random, shuffle
from typing import Dict, List

from pytz import all_timezones, timezone

from swh.model.hashutil import MultiHash

PROTOCOLS = ["git", "http", "https", "deb", "svn", "mock"]
DOMAINS = ["example.com", "some.long.host.name", "xn--n28h.tld"]
PATHS = [
    "",
    "/",
    "/stuff",
    "/stuff/",
    "/path/to/resource",
    "/path/with/anchor#id=42",
    "/path/with/qargs?q=1&b",
]
CONTENT_STATUS = ["visible", "hidden", "absent"]
MAX_DATE = 3e9  # around 2065


def gen_all_origins():
    for protocol in PROTOCOLS:
        for domain in DOMAINS:
            for urlpath in PATHS:
                yield {"url": "%s://%s%s" % (protocol, domain, urlpath)}


ORIGINS = list(gen_all_origins())


def gen_origins(n: int = 100) -> List:
    """Returns a list of n randomly generated origins suitable for using as
    Storage.add_origin() argument.

    """
    origins = ORIGINS[:]
    shuffle(origins)
    return origins[:n]


def gen_content():
    size = randint(1, 10 * 1024)
    data = bytes(randint(0, 255) for i in range(size))
    status = choice(CONTENT_STATUS)
    h = MultiHash.from_data(data)
    ctime = datetime.fromtimestamp(random() * MAX_DATE, timezone(choice(all_timezones)))
    content = {
        "data": data,
        "status": status,
        "length": size,
        "ctime": ctime,
        **h.digest(),
    }
    if status == "absent":
        content["reason"] = "why not"
        content["data"] = None
    return content


def gen_contents(n=20) -> List[Dict]:
    """Returns a list of n randomly generated content objects (as dict) suitable
    for using as Storage.content_add() argument.
    """
    return [gen_content() for i in range(n)]



# Copyright (C) 2015  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import unittest

from swh.model import exceptions, hashutil, validators


def hash_data(raw_content):
    return hashutil.MultiHash.from_data(raw_content).digest()


class TestValidators(unittest.TestCase):
    def setUp(self):
        self.valid_visible_content = {
            "status": "visible",
            "length": 5,
            "data": b"1984\n",
            "ctime": datetime.datetime(
                2015, 11, 22, 16, 33, 56, tzinfo=datetime.timezone.utc
            ),
        }

        self.valid_visible_content.update(hash_data(self.valid_visible_content["data"]))

        self.valid_absent_content = {
            "status": "absent",
            "length": 5,
            "ctime": datetime.datetime(
                2015, 11, 22, 16, 33, 56, tzinfo=datetime.timezone.utc
            ),
            "reason": "Content too large",
            "sha1_git": self.valid_visible_content["sha1_git"],
            "origin": 42,
        }

        self.invalid_content_hash_mismatch = self.valid_visible_content.copy()
        self.invalid_content_hash_mismatch.update(
            hash_data(b"this is not the data you're looking for")
        )

    def test_validate_content(self):
        self.assertTrue(validators.validate_content(self.valid_visible_content))

        self.assertTrue(validators.validate_content(self.valid_absent_content))

    def test_validate_content_hash_mismatch(self):
        with self.assertRaises(exceptions.ValidationError) as cm:
            validators.validate_content(self.invalid_content_hash_mismatch)

        # All the hashes are wrong. The exception should be of the form:
        # ValidationError({
        #     NON_FIELD_ERRORS: [
        #         ValidationError('content-hash-mismatch', 'sha1'),
        #         ValidationError('content-hash-mismatch', 'sha1_git'),
        #         ValidationError('content-hash-mismatch', 'sha256'),
        #     ]
        # })

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(set(exc.error_dict.keys()), {exceptions.NON_FIELD_ERRORS})

        hash_mismatches = exc.error_dict[exceptions.NON_FIELD_ERRORS]
        self.assertIsInstance(hash_mismatches, list)
        self.assertEqual(len(hash_mismatches), 4)
        self.assertTrue(
            all(
                mismatch.code == "content-hash-mismatch" for mismatch in hash_mismatches
            )
        )
        self.assertEqual(
            set(mismatch.params["hash"] for mismatch in hash_mismatches),
            {"sha1", "sha1_git", "sha256", "blake2s256"},
        )



# Copyright (C) 2015  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest

from swh.model.exceptions import ValidationError
from swh.model.fields import hashes


class ValidateHashes(unittest.TestCase):
    def setUp(self):
        self.valid_byte_hashes = {
            "sha1": b"\xf1\xd2\xd2\xf9\x24\xe9\x86\xac\x86\xfd\xf7\xb3\x6c\x94"
            b"\xbc\xdf\x32\xbe\xec\x15",
            "sha1_git": b"\x25\x7c\xc5\x64\x2c\xb1\xa0\x54\xf0\x8c\xc8\x3f\x2d"
            b"\x94\x3e\x56\xfd\x3e\xbe\x99",
            "sha256": b"\xb5\xbb\x9d\x80\x14\xa0\xf9\xb1\xd6\x1e\x21\xe7\x96"
            b"\xd7\x8d\xcc\xdf\x13\x52\xf2\x3c\xd3\x28\x12\xf4\x85"
            b"\x0b\x87\x8a\xe4\x94\x4c",
        }

        self.valid_str_hashes = {
            "sha1": "f1d2d2f924e986ac86fdf7b36c94bcdf32beec15",
            "sha1_git": "257cc5642cb1a054f08cc83f2d943e56fd3ebe99",
            "sha256": "b5bb9d8014a0f9b1d61e21e796d78dccdf1352f23cd32812f485"
            "0b878ae4944c",
        }

        self.bad_hash = object()

    def test_valid_bytes_hash(self):
        for hash_type, value in self.valid_byte_hashes.items():
            self.assertTrue(hashes.validate_hash(value, hash_type))

    def test_valid_str_hash(self):
        for hash_type, value in self.valid_str_hashes.items():
            self.assertTrue(hashes.validate_hash(value, hash_type))

    def test_invalid_hash_type(self):
        hash_type = "unknown_hash_type"
        with self.assertRaises(ValidationError) as cm:
            hashes.validate_hash(self.valid_str_hashes["sha1"], hash_type)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, "unexpected-hash-type")
        self.assertEqual(exc.params["hash_type"], hash_type)

        self.assertIn("Unexpected hash type", str(exc))
        self.assertIn(hash_type, str(exc))

    def test_invalid_bytes_len(self):
        for hash_type, value in self.valid_byte_hashes.items():
            value = value + b"\x00\x01"
            with self.assertRaises(ValidationError) as cm:
                hashes.validate_hash(value, hash_type)

            exc = cm.exception
            self.assertIsInstance(str(exc), str)
            self.assertEqual(exc.code, "unexpected-hash-length")
            self.assertEqual(exc.params["hash_type"], hash_type)
            self.assertEqual(exc.params["length"], len(value))

            self.assertIn("Unexpected length", str(exc))
            self.assertIn(str(len(value)), str(exc))

    def test_invalid_str_len(self):
        for hash_type, value in self.valid_str_hashes.items():
            value = value + "0001"
            with self.assertRaises(ValidationError) as cm:
                hashes.validate_hash(value, hash_type)

            exc = cm.exception
            self.assertIsInstance(str(exc), str)
            self.assertEqual(exc.code, "unexpected-hash-length")
            self.assertEqual(exc.params["hash_type"], hash_type)
            self.assertEqual(exc.params["length"], len(value))

            self.assertIn("Unexpected length", str(exc))
            self.assertIn(str(len(value)), str(exc))

    def test_invalid_str_contents(self):
        for hash_type, value in self.valid_str_hashes.items():
            value = "\xa2" + value[1:-1] + "\xc3"
            with self.assertRaises(ValidationError) as cm:
                hashes.validate_hash(value, hash_type)

            exc = cm.exception
            self.assertIsInstance(str(exc), str)
            self.assertEqual(exc.code, "unexpected-hash-contents")
            self.assertEqual(exc.params["hash_type"], hash_type)
            self.assertEqual(exc.params["unexpected_chars"], "\xa2, \xc3")

            self.assertIn("Unexpected characters", str(exc))
            self.assertIn("\xc3", str(exc))
            self.assertIn("\xa2", str(exc))

    def test_invalid_value_type(self):
        with self.assertRaises(ValidationError) as cm:
            hashes.validate_hash(self.bad_hash, "sha1")

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, "unexpected-hash-value-type")
        self.assertEqual(exc.params["type"], self.bad_hash.__class__.__name__)

        self.assertIn("Unexpected type", str(exc))
        self.assertIn(self.bad_hash.__class__.__name__, str(exc))

    def test_validate_sha1(self):
        self.assertTrue(hashes.validate_sha1(self.valid_byte_hashes["sha1"]))
        self.assertTrue(hashes.validate_sha1(self.valid_str_hashes["sha1"]))

        with self.assertRaises(ValidationError) as cm:
            hashes.validate_sha1(self.bad_hash)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, "unexpected-hash-value-type")
        self.assertEqual(exc.params["type"], self.bad_hash.__class__.__name__)

    def test_validate_sha1_git(self):
        self.assertTrue(hashes.validate_sha1_git(self.valid_byte_hashes["sha1_git"]))
        self.assertTrue(hashes.validate_sha1_git(self.valid_str_hashes["sha1_git"]))

        with self.assertRaises(ValidationError) as cm:
            hashes.validate_sha1_git(self.bad_hash)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, "unexpected-hash-value-type")
        self.assertEqual(exc.params["type"], self.bad_hash.__class__.__name__)

    def test_validate_sha256(self):
        self.assertTrue(hashes.validate_sha256(self.valid_byte_hashes["sha256"]))
        self.assertTrue(hashes.validate_sha256(self.valid_str_hashes["sha256"]))

        with self.assertRaises(ValidationError) as cm:
            hashes.validate_sha256(self.bad_hash)

        exc = cm.exception
        self.assertIsInstance(str(exc), str)
        self.assertEqual(exc.code, "unexpected-hash-value-type")
        self.assertEqual(exc.params["type"], self.bad_hash.__class__.__name__)



# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import sys
from typing import Dict, Iterable, Optional

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
try:
    import click
except ImportError:
    print(
        "Cannot run swh-identify; the Click package is not installed."
        "Please install 'swh.model[cli]' for full functionality.",
        file=sys.stderr,
    )
    exit(1)

try:
    from swh.core.cli import swh as swh_cli_group
except ImportError:
    # stub so that swh-identify can be used when swh-core isn't installed
    swh_cli_group = click  # type: ignore

from swh.model.from_disk import Directory
from swh.model.swhids import CoreSWHID

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

# Mapping between dulwich types and Software Heritage ones. Used by snapshot ID
# computation.
_DULWICH_TYPES = {
    b"blob": "content",
    b"tree": "directory",
    b"commit": "revision",
    b"tag": "release",
}


class CoreSWHIDParamType(click.ParamType):
    """Click argument that accepts a core SWHID and returns them as
    :class:`swh.model.swhids.CoreSWHID` instances """

    name = "SWHID"

    def convert(self, value, param, ctx) -> CoreSWHID:
        from swh.model.exceptions import ValidationError

        try:
            return CoreSWHID.from_string(value)
        except ValidationError as e:
            self.fail(f'"{value}" is not a valid core SWHID: {e}', param, ctx)


def swhid_of_file(path) -> CoreSWHID:
    from swh.model.from_disk import Content

    object = Content.from_file(path=path)
    return object.swhid()


def swhid_of_file_content(data) -> CoreSWHID:
    from swh.model.from_disk import Content

    object = Content.from_bytes(mode=644, data=data)
    return object.swhid()


def model_of_dir(path: bytes, exclude_patterns: Iterable[bytes] = None) -> Directory:
    from swh.model.from_disk import accept_all_directories, ignore_directories_patterns

    dir_filter = (
        ignore_directories_patterns(path, exclude_patterns)
        if exclude_patterns
        else accept_all_directories
    )

    return Directory.from_disk(path=path, dir_filter=dir_filter)


def swhid_of_dir(path: bytes, exclude_patterns: Iterable[bytes] = None) -> CoreSWHID:
    obj = model_of_dir(path, exclude_patterns)
    return obj.swhid()


def swhid_of_origin(url):
    from swh.model.model import Origin

    return Origin(url).swhid()


def swhid_of_git_repo(path) -> CoreSWHID:
    try:
        import dulwich.repo
    except ImportError:
        raise click.ClickException(
            "Cannot compute snapshot identifier; the Dulwich package is not installed. "
            "Please install 'swh.model[cli]' for full functionality.",
        )

    from swh.model import hashutil
    from swh.model.model import Snapshot

    repo = dulwich.repo.Repo(path)

    branches: Dict[bytes, Optional[Dict]] = {}
    for ref, target in repo.refs.as_dict().items():
        obj = repo[target]
        if obj:
            branches[ref] = {
                "target": hashutil.bytehex_to_hash(target),
                "target_type": _DULWICH_TYPES[obj.type_name],
            }
        else:
            branches[ref] = None

    for ref, target in repo.refs.get_symrefs().items():
        branches[ref] = {
            "target": target,
            "target_type": "alias",
        }

    snapshot = {"branches": branches}

    return Snapshot.from_dict(snapshot).swhid()


def identify_object(
    obj_type: str, follow_symlinks: bool, exclude_patterns: Iterable[bytes], obj
) -> str:
    from urllib.parse import urlparse

    if obj_type == "auto":
        if obj == "-" or os.path.isfile(obj):
            obj_type = "content"
        elif os.path.isdir(obj):
            obj_type = "directory"
        else:
            try:  # URL parsing
                if urlparse(obj).scheme:
                    obj_type = "origin"
                else:
                    raise ValueError
            except ValueError:
                raise click.BadParameter("cannot detect object type for %s" % obj)

    if obj == "-":
        content = sys.stdin.buffer.read()
        swhid = str(swhid_of_file_content(content))
    elif obj_type in ["content", "directory"]:
        path = obj.encode(sys.getfilesystemencoding())
        if follow_symlinks and os.path.islink(obj):
            path = os.path.realpath(obj)
        if obj_type == "content":
            swhid = str(swhid_of_file(path))
        elif obj_type == "directory":
            swhid = str(swhid_of_dir(path, exclude_patterns))
    elif obj_type == "origin":
        swhid = str(swhid_of_origin(obj))
    elif obj_type == "snapshot":
        swhid = str(swhid_of_git_repo(obj))
    else:  # shouldn't happen, due to option validation
        raise click.BadParameter("invalid object type: " + obj_type)

    # note: we return original obj instead of path here, to preserve user-given
    # file name in output
    return swhid


@swh_cli_group.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    "--dereference/--no-dereference",
    "follow_symlinks",
    default=True,
    help="follow (or not) symlinks for OBJECTS passed as arguments "
    + "(default: follow)",
)
@click.option(
    "--filename/--no-filename",
    "show_filename",
    default=True,
    help="show/hide file name (default: show)",
)
@click.option(
    "--type",
    "-t",
    "obj_type",
    default="auto",
    type=click.Choice(["auto", "content", "directory", "origin", "snapshot"]),
    help="type of object to identify (default: auto)",
)
@click.option(
    "--exclude",
    "-x",
    "exclude_patterns",
    metavar="PATTERN",
    multiple=True,
    help="Exclude directories using glob patterns \
    (e.g., ``*.git`` to exclude all .git directories)",
)
@click.option(
    "--verify",
    "-v",
    metavar="SWHID",
    type=CoreSWHIDParamType(),
    help="reference identifier to be compared with computed one",
)
@click.option(
    "-r", "--recursive", is_flag=True, help="compute SWHID recursively",
)
@click.argument("objects", nargs=-1, required=True)
def identify(
    obj_type,
    verify,
    show_filename,
    follow_symlinks,
    objects,
    exclude_patterns,
    recursive,
):
    """Compute the Software Heritage persistent identifier (SWHID) for the given
    source code object(s).

    For more details about SWHIDs see:

    \b
    https://docs.softwareheritage.org/devel/swh-model/persistent-identifiers.html

    Tip: you can pass "-" to identify the content of standard input.

    \b
    Examples::

    \b
      $ swh identify fork.c kmod.c sched/deadline.c
      swh:1:cnt:2e391c754ae730bd2d8520c2ab497c403220c6e3    fork.c
      swh:1:cnt:0277d1216f80ae1adeed84a686ed34c9b2931fc2    kmod.c
      swh:1:cnt:57b939c81bce5d06fa587df8915f05affbe22b82    sched/deadline.c

    \b
      $ swh identify --no-filename /usr/src/linux/kernel/
      swh:1:dir:f9f858a48d663b3809c9e2f336412717496202ab

    \b
      $ git clone --mirror https://forge.softwareheritage.org/source/helloworld.git
      $ swh identify --type snapshot helloworld.git/
      swh:1:snp:510aa88bdc517345d258c1fc2babcd0e1f905e93	helloworld.git

    """  # NoQA  # overlong lines in shell examples are fine
    from functools import partial
    import logging

    if exclude_patterns:
        exclude_patterns = set(pattern.encode() for pattern in exclude_patterns)

    if verify and len(objects) != 1:
        raise click.BadParameter("verification requires a single object")

    if recursive and not os.path.isdir(objects[0]):
        recursive = False
        logging.warn("recursive option disabled, input is not a directory object")

    if recursive:
        if verify:
            raise click.BadParameter(
                "verification of recursive object identification is not supported"
            )

        if not obj_type == ("auto" or "directory"):
            raise click.BadParameter(
                "recursive identification is supported only for directories"
            )

        path = os.fsencode(objects[0])
        dir_obj = model_of_dir(path, exclude_patterns)
        for sub_obj in dir_obj.iter_tree():
            path_name = "path" if "path" in sub_obj.data.keys() else "data"
            path = os.fsdecode(sub_obj.data[path_name])
            swhid = str(sub_obj.swhid())
            msg = f"{swhid}\t{path}" if show_filename else f"{swhid}"
            click.echo(msg)
    else:
        results = zip(
            objects,
            map(
                partial(identify_object, obj_type, follow_symlinks, exclude_patterns),
                objects,
            ),
        )

        if verify:
            swhid = next(results)[1]
            if str(verify) == swhid:
                click.echo("SWHID match: %s" % swhid)
                sys.exit(0)
            else:
                click.echo("SWHID mismatch: %s != %s" % (verify, swhid))
                sys.exit(1)
        else:
            for (obj, swhid) in results:
                msg = swhid
                if show_filename:
                    msg = "%s\t%s" % (swhid, os.fsdecode(obj))
                click.echo(msg)


if __name__ == "__main__":
    identify()



# Copyright (C) 2015  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import string

from ..exceptions import ValidationError


def validate_hash(value, hash_type):
    """Validate that the given value represents a hash of the given hash_type.

    Args:
        value: the value to check
        hash_type: the type of hash the value is representing

    Returns:
        True if the hash validates

    Raises:
        ValueError if the hash does not validate
    """

    hash_lengths = {
        "sha1": 20,
        "sha1_git": 20,
        "sha256": 32,
    }

    hex_digits = set(string.hexdigits)

    if hash_type not in hash_lengths:
        raise ValidationError(
            "Unexpected hash type %(hash_type)s, expected one of" " %(hash_types)s",
            params={
                "hash_type": hash_type,
                "hash_types": ", ".join(sorted(hash_lengths)),
            },
            code="unexpected-hash-type",
        )

    if isinstance(value, str):
        errors = []
        extra_chars = set(value) - hex_digits
        if extra_chars:
            errors.append(
                ValidationError(
                    "Unexpected characters `%(unexpected_chars)s' for hash "
                    "type %(hash_type)s",
                    params={
                        "unexpected_chars": ", ".join(sorted(extra_chars)),
                        "hash_type": hash_type,
                    },
                    code="unexpected-hash-contents",
                )
            )

        length = len(value)
        expected_length = 2 * hash_lengths[hash_type]
        if length != expected_length:
            errors.append(
                ValidationError(
                    "Unexpected length %(length)d for hash type "
                    "%(hash_type)s, expected %(expected_length)d",
                    params={
                        "length": length,
                        "expected_length": expected_length,
                        "hash_type": hash_type,
                    },
                    code="unexpected-hash-length",
                )
            )

        if errors:
            raise ValidationError(errors)

        return True

    if isinstance(value, bytes):
        length = len(value)
        expected_length = hash_lengths[hash_type]
        if length != expected_length:
            raise ValidationError(
                "Unexpected length %(length)d for hash type "
                "%(hash_type)s, expected %(expected_length)d",
                params={
                    "length": length,
                    "expected_length": expected_length,
                    "hash_type": hash_type,
                },
                code="unexpected-hash-length",
            )

        return True

    raise ValidationError(
        "Unexpected type %(type)s for hash, expected str or bytes",
        params={"type": value.__class__.__name__,},
        code="unexpected-hash-value-type",
    )


def validate_sha1(sha1):
    """Validate that sha1 is a valid sha1 hash"""
    return validate_hash(sha1, "sha1")


def validate_sha1_git(sha1_git):
    """Validate that sha1_git is a valid sha1_git hash"""
    return validate_hash(sha1_git, "sha1_git")


def validate_sha256(sha256):
    """Validate that sha256 is a valid sha256 hash"""
    return validate_hash(sha256, "sha256")
